//
//  CommandLineInterface.cpp
//  LevelSetMethod2
//
//  Created by kumada on 2012/12/18.
//  Copyright (c) 2012年 kumada. All rights reserved.
//

#include "CommandLineInterface.h"
#include "LevelSetMethodViewer2d.h"
#include "LevelSetMethodViewer3d.h"
#include "GeometryGenerator.h"

#include <iostream>
#include <string>
#include <iomanip>
#include <fstream>

#include <boost/program_options.hpp>
#include <opencv2/opencv.hpp>

namespace program_options = boost::program_options;
using namespace lsm;

namespace
{
        /// extract a parameter from "vm"
        /**
         *      @param[in] name          parameter name to extract
         *      @param[in] error_message error message in case of failure
         *      @param[in] vm            argument container
         */
        template<typename T>
        inline const T extract_parameter(
                const std::string&                    name,
                const std::string&                    error_message,
                const program_options::variables_map& vm
        ) {
                return vm.count(name) ? vm[name].as<T>() : throw std::runtime_error(error_message);
        }
        
        /// load an input image used for the Level Set Method in 2D
        /**
         *      @param[in]  filename   input file name
         *      @param[out] image      input image
         *      @param[out] space_size image size
         */
        void load_input_image(
                const std::string&         filename,
                std::vector<std::uint8_t>& image,
                SpaceSize<TwoDimension>&   space_size
        ) {
                cv::Mat gray = cv::imread(filename, CV_8UC1);
                if ( gray.empty() ) {
                        throw std::runtime_error("unable to load an input image");
                }
                space_size.width_ = gray.cols;
                space_size.height_ = gray.rows;
                space_size.total_ = gray.cols * gray.rows;
                std::copy(gray.data, gray.data + space_size.total_, std::back_inserter(image));
        }
        
        void set_common_arguments(program_options::options_description& desc)
        {
                desc.add_options()
                        ("help",                                                   "produce help message")
                        ("dim",             program_options::value<int>(),         "set either 2 or 3")
                        ("verbose",                                                "print verbose description")
                        ("input",           program_options::value<std::string>(), "set GRAY image path in 2D/set pattern in 3D")
                        ("wband",           program_options::value<int>(),         "set width of band")
                        ("wreset",          program_options::value<int>(),         "set width to reset")
                        ("time_step",       program_options::value<double>(),      "set time step")
                        ("gain",            program_options::value<double>(),      "set gain")
                        ("constant_speed",  program_options::value<double>(),      "set constant speed")
                        ("speed_threshold", program_options::value<double>(),      "set speed threshold")
                        ("left",            program_options::value<int>(),         "set left of initial rectangle")
                        ("top",             program_options::value<int>(),         "set top of initial rectangle")
                        ("right",           program_options::value<int>(),         "set right of initial rectangle")
                        ("bottom",          program_options::value<int>(),         "set bottom of initial rectangle")
                        ("front",           program_options::value<int>(),         "set front of initial rectangle (not used in 2D)")
                        ("back",            program_options::value<int>(),         "set back of initial rectangle (not used in 2D)")
                ;
        }
        
        void set_parameters(const program_options::variables_map& vm, Parameters& params)
        {
                params.wband_           = extract_parameter<int>(   "wband",           "invalid wband",           vm);
                params.wreset_          = extract_parameter<int>(   "wreset",          "invalid wreset",          vm);
                params.time_step_       = extract_parameter<double>("time_step",       "invalid time step",       vm);
                params.gain_            = extract_parameter<double>("gain",            "invalid gain",            vm);
                params.constant_speed_  = extract_parameter<double>("constant_speed",  "invalid constant speed",  vm);
                params.speed_threshold_ = extract_parameter<double>("speed_threshold", "invalid speed threshold", vm);
        }
}

std::unique_ptr<LevelSetMethodViewer2d> CommandLineInterface::viewer2d_;

void CommandLineInterface::execute_level_set_method(int argc, char* argv[])
{
        program_options::options_description desc("allowed options");
        set_common_arguments(desc);

        program_options::variables_map vm;
        store(parse_command_line(argc, argv, desc), vm);
        notify(vm);

        if ( vm.empty() || vm.count("help") ) {
                std::cout << desc << "\n";
                return;
        }

        Parameters params;
        set_parameters(vm, params);
        
        int dim = extract_parameter<int>("dim", "invalid dimension", vm);
        if ( dim == 2 ) {
                execute_level_set_method_in_2d(argc, argv, vm, params);
        } else {
                execute_level_set_method_in_3d(argc, argv, vm, params);
        }
}

void CommandLineInterface::display_2d()
{
        viewer2d_->display();
}

void CommandLineInterface::keyboard_2d(unsigned char key, int x, int y)
{
        viewer2d_->keyboard(key, x, y);
}

/// execute the Level Set Method in 2D
void CommandLineInterface::execute_level_set_method_in_2d(
        int                                   argc,
        char*                                 argv[],
        const program_options::variables_map& vm,
        const Parameters&                     params
) {
        // set an initial front
        const int left   = extract_parameter<int>("left",   "invalid left",   vm);
        const int top    = extract_parameter<int>("top",    "invalid top",    vm);
        const int right  = extract_parameter<int>("right",  "invalid left",   vm);
        const int bottom = extract_parameter<int>("bottom", "invalid bottom", vm);
        InitialFront<TwoDimension> initial_front;
        initial_front.vertices_[0] = {{left, top}};
        initial_front.vertices_[1] = {{right, bottom}};
 
        // load an input image
        const std::string filename = extract_parameter<std::string>("input", "invalid input", vm);
        SpaceSize2d space_size;
        std::vector<std::uint8_t> pixels;
        load_input_image(filename, pixels, space_size);
        
        viewer2d_.reset(new LevelSetMethodViewer2d);
        viewer2d_->set_texture(&pixels[0]);
        
        if ( vm.count("verbose") ) {
                viewer2d_->set_verbose(true);
        }
        
        viewer2d_->initialize_viewer(argc, argv, space_size, "Level Set Method in 2D", display_2d, keyboard_2d);
        viewer2d_->initialize_level_set_method(params, space_size, &pixels[0], initial_front);
        viewer2d_->start();
}

namespace
{
        enum class Pattern
        {
                Pattern0,
                Pattern1,
        };
        
        /// create 3D objects
        void create_objects(
                Pattern                    pattern,
                const Parameters&          params,
                GeometryGenerator&         geometry_generator,
                int                        factor,
                std::vector<std::uint8_t>& src
        ) {
                const auto& space_size = geometry_generator.get_space_size();
                const int w = space_size.width_;
                const int h = space_size.height_;
                const int d = space_size.depth_;

                switch ( pattern ) {
                case Pattern::Pattern0:
                {
                        const int radius {15 * factor};
                        bool is_solid {true};
                        const Sphere sphere0 { {{w / 4, h / 4, d / 4}}, radius, is_solid };
                        const Sphere sphere1 { {{3 * w / 4, 3 * h / 4, 3 * d / 4}}, radius, is_solid };
                        geometry_generator.add_sphere(sphere0);
                        geometry_generator.add_sphere(sphere1);
                        geometry_generator.generate(src);
                        break;
                }

                case Pattern::Pattern1:
                {
                        const int radius {15 * factor};
                        bool is_solid {true};
                        const Sphere sphere0 { {{w / 2, h / 2, d / 4}}, radius, is_solid };
                        const Sphere sphere1 { {{w / 2, 3 * h / 4, d / 2}}, radius, is_solid };
                        const Sphere sphere2 { {{w / 2, h / 2, d / 2}}, std::min(w, std::min(h, d)) / 2 - params.wband_, !is_solid };
                        geometry_generator.add_sphere(sphere0);
                        geometry_generator.add_sphere(sphere1);
                        geometry_generator.add_sphere(sphere2);
                        geometry_generator.generate(src);
                        break;
                }

                default:
                        break;
                }
        }
        
        inline const Pattern extract_pattern(const std::string& input)
        {
                if ( input == "pattern0" ) {
                        return Pattern::Pattern0;
                }
                if ( input == "pattern1" ) {
                        return  Pattern::Pattern1;
                }
                throw std::runtime_error("invalid pattern");
        }
}

std::unique_ptr<LevelSetMethodViewer3d> CommandLineInterface::viewer3d_;

void CommandLineInterface::display_3d()
{
        viewer3d_->display();
}

void CommandLineInterface::keyboard_3d(unsigned char key, int x, int y)
{
        viewer3d_->keyboard(key, x, y);
}

/// execute the Level Set Method in 3D
void CommandLineInterface::execute_level_set_method_in_3d(
        int                                   argc,
        char*                                 argv[],
        const program_options::variables_map& vm,
        const Parameters&                     params
) {
        // set an initial front
        const int left   = extract_parameter<int>("left",   "invalid left",   vm);
        const int top    = extract_parameter<int>("top",    "invalid top",    vm);
        const int right  = extract_parameter<int>("right",  "invalid left",   vm);
        const int bottom = extract_parameter<int>("bottom", "invalid bottom", vm);
        const int front  = extract_parameter<int>("front",  "invalid front",  vm);
        const int back   = extract_parameter<int>("back",   "invalid back",   vm);
        InitialFront<ThreeDimension> initial_front;
        initial_front.vertices_[0] = {{left, top, front}};
        initial_front.vertices_[1] = {{right, bottom, back}};
        
        // set a 3D-space size
        constexpr int factor = 2;
        constexpr int width  = factor * 100;
        constexpr int height = factor * 100;
        constexpr int depth  = factor * 100;
        SpaceSize<ThreeDimension> space_size {width, height, depth};

        // create 3D objects
        std::vector<std::uint8_t> pixels(space_size.total_, 255);
        GeometryGenerator geometry_generator {space_size};
        const std::string input = extract_parameter<std::string>("input", "invalid input", vm);
        create_objects(extract_pattern(input), params, geometry_generator, factor, pixels);
        
        viewer3d_.reset(new LevelSetMethodViewer3d);
        const auto& spheres = geometry_generator.get_spheres();
        for ( const auto& sphere : spheres ) {
                viewer3d_->add_sphere(sphere);
        }

        if ( vm.count("verbose") ) {
                viewer3d_->set_verbose(true);
        }
        
        viewer3d_->initialize_viewer(argc, argv, space_size, "Level Set Method in 3D", display_3d, keyboard_3d);
        viewer3d_->initialize_level_set_method(params, space_size, &pixels[0], initial_front);
        viewer3d_->start();
}